@@ -3,6 +3,7 @@ wia_scan/scan_functions.py: an accumulation of helper functions
 """
 
 import os
+import tempfile
 from datetime import datetime
 
 import win32com.client
@@ -35,7 +36,7 @@ WIA_EXTENSION_TO_FORMAT = {
 }
 
 
-DEFAULT_SCAN_SETTINGS = {
+DEFAULT_SCAN_PROFILE = {
     'brightness': 0,
     'contrast': 0,
     'dpi': 200,
@@ -61,7 +62,10 @@ class IndentPrinter:
         self.print_function(output)
 
 
-def print_device_properties(device, print_function):
+DEFAULT_PRINT_FUNCTION = IndentPrinter(indent=0, print_function=print)
+
+
+def print_device_properties(device, print_function=DEFAULT_PRINT_FUNCTION):
     print_function('Device Commands: ', indent=0)
     for command in device.Commands:
         print_function(f'{command.Name}: {command.Description}', indent=1)
@@ -82,14 +86,14 @@ def press_any_key_to_continue():
     os.system("pause")
 
 
-def print_object_debug(name, obj, print_function):
+def print_object_debug(name, obj, print_function=DEFAULT_PRINT_FUNCTION):
     print_function(name, ' ', type(obj))
     for property_ in dir(obj):
         print_function('obj.property_ ', property_, indent=1)
     print_function(obj)
 
 
-def print_wia_image_debug(wia_image, print_function):
+def print_wia_image_debug(wia_image, print_function=DEFAULT_PRINT_FUNCTION):
     print_object_debug('image ', wia_image,
                        print_function=print_function)
     print_function('format id', wia_image.FormatID)
@@ -120,7 +124,7 @@ def get_device_manager():
     return device_manager
 
 
-def list_devices(device_manager, print_function, verbose=False):
+def list_devices(device_manager, print_function=DEFAULT_PRINT_FUNCTION, verbose=False):
     """ Prints the available devices to print_function """
 
     if device_manager.DeviceInfos.Count == 0:
@@ -153,7 +157,7 @@ def ask_for_an_int(prompt_message, default, valid_range):
             print('Could not understand answer, please try again')
 
 
-def prompt_choose_device(device_manager, print_function, verbose=False):
+def prompt_choose_device(device_manager, print_function=DEFAULT_PRINT_FUNCTION, verbose=False):
     if device_manager.DeviceInfos.Count == 0:
         raise ValueError(
             "devices must be available to choose, but none is available")
@@ -186,8 +190,38 @@ def get_device_info_by_unique_id(device_manager, unique_id):
     return None
 
 
-def connect_device(device_manager, device_identifier, print_function, verbose=False, quiet=False):
-    print_function('Entered device identifier', device_identifier)
+def print_item_tree(item, print_function=DEFAULT_PRINT_FUNCTION):
+    print_device_properties(item, print_function=print_function)
+    more_indent_function = IndentPrinter(
+        indent=1, print_function=print_function)
+    for item_index in range(1, item.Items.Count + 1):
+        print_function('sub item ', item_index)
+        item_ = item.Items(item_index)
+        print_item_tree(item_, print_function=more_indent_function)
+
+
+def get_scan_source(device, print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
+    if device.Items.Count == 1:
+        device = device.Items(1)
+    elif device.Items.Count == 0:
+        pass
+    else:
+        raise NotImplementedError(
+            'only one scan source is implemented at this point')
+
+    assert device.Items.Count == 0
+    if verbose:
+        print_object_debug('device ', device,
+                           print_function=print_function)
+        print_device_properties(device, print_function=print_function)
+
+    return device
+
+
+def connect_device(device_manager, device_identifier,
+                   print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
+    if device_identifier is None:
+        raise ValueError('device_identifier cannot be None')
     device_info = get_device_info_by_unique_id(
         device_manager, unique_id=device_identifier)
     if device_info is None:
@@ -199,26 +233,37 @@ def connect_device(device_manager, device_identifier, print_function, verbose=Fa
     if not quiet:
         print_function('Connected ')
 
-    assert device.Items.Count <= 1
-    if device.Items.Count == 1:  # has sub item, i have no idea why though
-        if verbose:
-            print_function('Getting subdevice')
-        device = device.Items(1)
-        if verbose:
-            print_object_debug('device ', device,
-                               print_function=print_function)
-            print_device_properties(device, print_function=print_function)
+    if verbose:
+        print_function('Printing item tree: ')
+        print_item_tree(device, print_function=print_function)
+
+    device = get_scan_source(
+        device, print_function=print_function, verbose=verbose, quiet=quiet)
 
     return device
 
 
-def set_device_property(device, property_name, value):
-    for property_ in device.Properties:
-        if property_.Name == property_name:
-            property_.Value = value
+def connect_to_device_by_uid(
+        device_uid, print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
+    device_manager = get_device_manager()
+    device = connect_device(device_manager, device_uid,
+                            print_function=print_function, verbose=verbose, quiet=quiet)
+    return device
 
 
-def scan_side(device, settings, print_function, verbose=False, quiet=False):
+def prompt_choose_device_and_connect(
+        print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
+    device_manager = get_device_manager()
+    device_identifier = prompt_choose_device(
+        device_manager=device_manager, print_function=print_function, verbose=verbose)
+    device = connect_device(device_manager=device_manager,
+                            device_identifier=device_identifier, print_function=print_function,
+                            verbose=verbose, quiet=quiet)
+    return device
+
+
+def scan_side(device, scan_profile, return_wia_image=False,
+              print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
     """
     Scans a single side and returns the WIA Image
 
@@ -226,7 +271,7 @@ def scan_side(device, settings, print_function, verbose=False, quiet=False):
     ----------
     device : wia device
         can be found by calling connect_device
-    settings : dict{
+    scan_profile : dict{
         'brightness': 0   Goes from -1000 to 1000
         'contrast': 0     Goes from -1000 to 1000
         'dpi': 200        Supported values depend on scanner
@@ -235,14 +280,14 @@ def scan_side(device, settings, print_function, verbose=False, quiet=False):
 
     """
 
-    set_device_property(device, 'Brightness', settings['brightness'])
-    set_device_property(device, 'Contrast', settings['contrast'])
-    set_device_property(device, 'Horizontal Resolution', settings['dpi'])
-    set_device_property(device, 'Vertical Resolution', settings['dpi'])
-    if settings['mode'] == 'RGB':
-        set_device_property(device, 'Data Type', 3)
-    elif settings['mode'] == 'L':
-        set_device_property(device, 'Data Type', 2)
+    device.Properties('Brightness').Value = scan_profile['brightness']
+    device.Properties('Contrast').Value = scan_profile['contrast']
+    device.Properties('Horizontal Resolution').Value = scan_profile['dpi']
+    device.Properties('Vertical Resolution').Value = scan_profile['dpi']
+    if scan_profile['mode'] == 'RGB':
+        device.Properties('Data Type').Value = 3
+    elif scan_profile['mode'] == 'L':
+        device.Properties('Data Type').Value = 2
     else:
         raise ValueError('unsupported mode')
 
@@ -261,33 +306,120 @@ def scan_side(device, settings, print_function, verbose=False, quiet=False):
         if command.CommandID == WIA_COMMAND_TAKE_PICTURE:
             device.ExecuteCommand(WIA_COMMAND_TAKE_PICTURE)
 
-    # Warning: can return a different format since item.Formats returns actually
-    # supported formats for transfer
-    image = device.Transfer(WIA_FORMAT_BMP)
+    wia_image = device.Transfer(WIA_FORMAT_BMP)
+    # if bmp is not supported as transfer format, wia will automatically choose the preferred method
+    # https://learn.microsoft.com/en-us/previous-versions/windows/desktop/wiaaut/-wiaaut-iitem-transfer
 
     if not quiet:
         print_function('Done...')
 
-    return image
+    if verbose:
+        print_wia_image_debug(wia_image, print_function=print_function)
+
+    if return_wia_image:
+        return wia_image
+
+    # to convert to a pillow image, we save it to disk, this is much faster on most current platforms
+    # then reading the ARGB data from wia_image, as long as we do not want to
+    # add numba or sth as a dependency
+
+    tmp_directory = tempfile.mkdtemp()
+    if verbose:
+        print_function(f'Created temporary directory {tmp_directory}')
+    tmp_file_name = os.path.join(tmp_directory, 'tmp.png')
+
+    if verbose:
+        print_function(f'Saving temporary image to {tmp_file_name}')
+
+    wia_image.SaveFile(tmp_file_name)
+
+    pillow_image_file = PILImage.open(tmp_file_name)
+    pillow_image = pillow_image_file.copy()
+    pillow_image_file.close()
+
+    os.remove(tmp_file_name)
+    os.rmdir(tmp_directory)
+
+    return pillow_image
 
 
-def scan_single_side_main(print_function, verbose):
-    device_manager = get_device_manager()
-    device_identifier = prompt_choose_device(
-        device_manager=device_manager, print_function=print_function, verbose=verbose)
-    device = connect_device(device_manager=device_manager,
-                            device_identifier=device_identifier, print_function=print_function, verbose=verbose)
-    wia_image = scan_side(device=device, settings=DEFAULT_SCAN_SETTINGS,
-                          print_function=print_function, verbose=verbose)
-
-    now = datetime.now()
-    date_string = now.strftime("%Y-%m-%d_%H-%M-%S")
-    filename = f'scan_{date_string}.png'
-
-    wia_image.SaveFile(filename)
+def get_device_quiet_or_choose(device_uid, print_function, verbose, quiet):
+    if quiet:
+        if device_uid is None:
+            raise ValueError('device_uid cannot be None when quiet')
+        device = connect_to_device_by_uid(
+            device_uid, print_function=print_function, verbose=verbose, quiet=quiet)
+    else:
+        device = prompt_choose_device_and_connect(
+            print_function=print_function, verbose=verbose, quiet=quiet)
+    return device
 
 
-def prompt_scan_profile(print_function):
+def run_calibration_process(
+        start_range, end_range, num_runs, device_uid, output_folder,
+        print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False):
+    assert num_runs > 1
+
+    if not quiet:
+        print_function(
+            'This calibration process consists of running a few scans')
+        print_function(
+            ' and you choosing yourself visually which setting is the best')
+        press_any_key_to_continue()
+
+    device = get_device_quiet_or_choose(
+        device_uid=device_uid, print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=quiet)
+
+    if not quiet:
+        print_function(
+            'Please put a calibration piece of paper in the scanner')
+        print_function(
+            'Consider something like a print of an it8 target (unrelated to this project)')
+        press_any_key_to_continue()
+
+    for i in range(num_runs):
+        lam_ = i / float(num_runs - 1)
+        brightness = (1.0 - lam_) * start_range + lam_ * end_range
+        scan_profile = {
+            'brightness': brightness,
+            'contrast': 0,
+            'dpi': 100,
+            'mode': 'RGB'
+        }
+
+        pillow_image = scan_side(
+            device=device,
+            scan_profile=scan_profile,
+            print_function=print_function,
+            verbose=verbose,
+            quiet=quiet)
+        filepath = os.path.join(
+            output_folder, f'run={i+1}_brightness={brightness}.jpg')
+        pillow_image.save(filepath, subsampling=0,
+                          optimize=True, progressive=True, quality=80)
+
+
+def scan_single_side_main(
+        scan_profile=None, print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=False,
+        device_uid=None, output_file=None):
+    if scan_profile is None:
+        scan_profile = DEFAULT_SCAN_PROFILE
+
+    device = get_device_quiet_or_choose(
+        device_uid=device_uid, print_function=DEFAULT_PRINT_FUNCTION, verbose=False, quiet=quiet)
+
+    wia_image = scan_side(device=device, scan_profile=scan_profile, return_wia_image=True,
+                          print_function=print_function, verbose=verbose, quiet=quiet)
+
+    if output_file is None:
+        now = datetime.now()
+        date_string = now.strftime("%Y-%m-%d_%H-%M-%S")
+        output_file = f'scan_{date_string}.png'
+
+    wia_image.SaveFile(output_file)
+
+
+def prompt_scan_profile(print_function=DEFAULT_PRINT_FUNCTION):
     while True:
         default_quality = 'g'
         quality = input(
@@ -309,46 +441,59 @@ def get_profile(profile_identifier):
         mode
         dpi
         jpeg_quality
+        brightness
+
     """
     if profile_identifier == 'h':
         return {
             'file_extension': 'png',
             'mode': 'RGB',
             'dpi': 300,
-            'jpeg_quality': 100
+            'jpeg_quality': 100,
+            'brightness': 0,
+            'contrast': 0,
         }
     if profile_identifier == 'c':
         return {
             'file_extension': 'jpg',
             'mode': 'RGB',
             'dpi': 200,
-            'jpeg_quality': 75
+            'jpeg_quality': 75,
+            'brightness': 0,
+            'contrast': 0,
         }
     if profile_identifier == 'g':
         return {
             'file_extension': 'jpg',
             'mode': 'L',
             'dpi': 200,
-            'jpeg_quality': 75
+            'jpeg_quality': 75,
+            'brightness': 0,
+            'contrast': 0,
         }
     raise NotImplementedError
 
 
-def scan_many_documents_flatbed(print_function, verbose=False):
-    device_manager = get_device_manager()
-    device_identifier = prompt_choose_device(
-        device_manager=device_manager, print_function=print_function, verbose=verbose)
-    device = connect_device(device_manager=device_manager,
-                            device_identifier=device_identifier,
-                            print_function=print_function,
-                            verbose=verbose)
+def scan_many_documents_flatbed(print_function=DEFAULT_PRINT_FUNCTION,
+                                verbose=False, output_folder=None):
+    if output_folder is None:
+        output_folder = '.'
+    else:
+        if not os.path.isdir(output_folder):
+            raise ValueError(f'Error: {output_folder} is not a directory')
+
+    quiet = False
+    device = prompt_choose_device_and_connect(
+        print_function=print_function, verbose=verbose, quiet=quiet)
 
     while True:
         profile_identifier = prompt_scan_profile(print_function=print_function)
         profile = get_profile(profile_identifier)
 
         number_of_sides = ask_for_an_int(
-            'Input number of sides of this document ', default=1, valid_range=(1, 10000))
+            'Input number of sides of this document (0 to stop)', default=1, valid_range=(1, 10000))
+        if number_of_sides == 0:
+            break
 
         pillow_image_combined = None
         pillow_image_side_height = None
@@ -357,50 +502,36 @@ def scan_many_documents_flatbed(print_function, verbose=False):
                 print_function('Please flip page')
                 press_any_key_to_continue()
             print_function(f'Scanning side {side+1} out of {number_of_sides}')
-            settings = {
-                'brightness': -150,
-                'contrast': 0,
+            scan_profile = {
+                'brightness': profile['brightness'],
+                'contrast': profile['contrast'],
                 'dpi': profile['dpi'],
-                'mode': profile['mode']
+                'mode': profile['mode'],
             }
-            wia_image = scan_side(
-                device=device, settings=settings, print_function=print_function,
-                verbose=verbose)
+            pillow_image = scan_side(
+                device=device, scan_profile=scan_profile, print_function=print_function,
+                verbose=verbose, quiet=quiet)
 
-            if verbose:
-                print_wia_image_debug(wia_image, print_function=print_function)
-
-            tmp_file_name = 'tmp.png'
-
-            if os.path.exists(tmp_file_name):
-                os.remove(tmp_file_name)
-            wia_image.SaveFile(tmp_file_name)
-
-            with PILImage.open(tmp_file_name) as pillow_image:
-                if pillow_image_combined is None:
-                    total_width = pillow_image.size[0]
-                    pillow_image_side_height = pillow_image.size[1]
-                    total_height = pillow_image_side_height * number_of_sides
-                    pillow_image_combined = PILImage.new(
-                        profile['mode'], (total_width, total_height))
-                # expecting all scans to give the same resolution...
-                assert pillow_image.height == pillow_image_side_height
-                y_offset = side * pillow_image_side_height
-                pillow_image_combined.paste(pillow_image, (0, y_offset))
-                pillow_image.close()
-
-            os.remove(tmp_file_name)
-
-        # if pillow_mode == 'L':
-        #   pillow_image_combined = pillow_image_combined.convert('L')
+            if pillow_image_combined is None:
+                total_width = pillow_image.size[0]
+                pillow_image_side_height = pillow_image.size[1]
+                total_height = pillow_image_side_height * number_of_sides
+                pillow_image_combined = PILImage.new(
+                    profile['mode'], (total_width, total_height))
+            # expecting all scans to give the same resolution...
+            assert pillow_image.height == pillow_image_side_height
+            y_offset = side * pillow_image_side_height
+            pillow_image_combined.paste(pillow_image, (0, y_offset))
 
         now = datetime.now()
         date_string = now.strftime("%Y-%m-%d_%H-%M-%S")
         filename = f"scan_{date_string}.{profile['file_extension']}"
+        filepath = os.path.join(output_folder, filename)
+
         if profile['file_extension'] == 'jpg':
-            pillow_image_combined.save(filename, subsampling=0, optimize=True,
+            pillow_image_combined.save(filepath, subsampling=0, optimize=True,
                                        progressive=True, quality=profile['jpeg_quality'])
         elif profile['file_extension'] == 'png':
-            pillow_image_combined.save(filename, optimize=True)
+            pillow_image_combined.save(filepath, optimize=True)
         else:
             raise ValueError("unexpected extension")
